#=======================================================================
#	$Id: Form.pm,v 1.3 2005/09/07 09:55:15 pythontech Exp $
#	Web form-filling
#	Copyright (C) 2005  Python Technology Limited
#
#	This program is free software; you can redistribute it and/or
#	modify it under the terms of the GNU General Public License
#	as published by the Free Software Foundation; either version 2
#	of the License, or (at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  
#	02111-1307, USA.
#
#	$form = new PythonTech::Form;
#	$form->add_text('comment',
#			initial => 'Hallo world',
#			validate => $re_or_code);
#	$form->add_int('count,
#		       initial => 42,
#		       min => 1, max => 100);
#	NOTYET $form->add_select('colour',
#			  list => [qw(red yellow blue)],
#			  initial => 'red',
#			  show => $hash_or_code);
#	$form->add_select_obj('user'
#			      list => [@all_users],
#			      initial => $thisuser,
#			      show => $attr_or_code);
#	$form->add_date('date',
#			initial => '2005-03-27');
#
#	if ($form->done($query)) {...}
#
#	my @nvlist = $form->values;
#
#	Filling form from template:
#
#	<input type="text" name="comment" size="40" value="{comment}" />
#
#	<input type="text" name="count" size="10" value="{count}" />
#
#	<select name="user">
#	{LOOP user}
#	<option value="/id"{/selected}>{/show}</option>
#	{ENDLOOP}
#
#	<input type="text" name="date" size="10" value="{date}" />
#	  or
#	<input type="text" name="date/dd" size="2" value="{date/dd}" />
#	<input type="text" name="date/mm" size="2" value="{date/mm}" />
#	<input type="text" name="date/yyyy" size="2" value="{date/yyyy}" />
#	  or
#	<select name="date/dd">
#	<option value="01"{date/dd/selected:01}> 1</option>
#	...
#	<option value="31"{date/dd/selected:31}> 1</option>
#	</select>
#	<select name="date/mm">
#	<option value="01"{date/mm/selected:01}>January</option>
#	...
#	<option value="12"{date/mm/selected:12}>December</option>
#	</select>
#	<select name="date/yyyy">
#	<option value="2004"{date/yyyy/selected:2004}>2004</option>
#	...
#	</select>
#=======================================================================
package PythonTech::Form;
use strict;
use Exporter;

our @ISA = qw(Exporter);
our @EXPORT = qw(form_text form_select_obj form_date);

sub new {
    my $class = shift;
    my $self = {fields=>{}, rx=>undef, @_};
    bless $self, $class;
    return $self;
}

#-----------------------------------------------------------------------
#	Methods for adding fields
#-----------------------------------------------------------------------
sub add_text {
    my($self,$name,@attr) = @_;
    my $fld = PythonTech::Form::Text->new(name => $name,
					  form => $self,
					  @attr);
    $self->{'fields'}->{$name} = $fld;
    $self->{'rx'} = undef;
    return $self;
}

sub add_int {
    my($self,$name,@attr) = @_;
    my $fld = PythonTech::Form::Text->new(name => $name,
					  form => $self,
					  validate => '(\d+)',
					  @attr);
    $self->{'fields'}->{$name} = $fld;
    $self->{'rx'} = undef;
    return $self;
}

sub add_date {
    my($self,$name,@attr) = @_;
    my $fld = PythonTech::Form::Date->new(name => $name,
					  form => $self,
					  @attr);
    $self->{'fields'}->{$name} = $fld;
    $self->{'rx'} = undef;
    return $self;
}

sub add_select_obj {
    my($self,$name,@attr) = @_;
    my $fld = PythonTech::Form::SelectObj->new(name => $name,
					       form => $self,
					       @attr);
    $self->{'fields'}->{$name} = $fld;
    $self->{'rx'} = undef;
    return $self;
}

#-----------------------------------------------------------------------
#	Apply query to the form and check if the user has successfully
#	entered valid data.  If the query is a GET, then any parameters
#	are used solely to override the default initial values.  If the
#	query is a POST, then assume the user is submitting the form
#	and check for errors.
#
#	Parameter values, or fragents thereof (e.g. year, month and day
#	of a date may be supplied separately) are presented to the
#	form fields via their setnew method, followed by a endnew.
#	The check method (called only for a POST) prompts each field to
#	update its error status.
#-----------------------------------------------------------------------
sub consider {
    my($self, $query) = @_;
    foreach my $var ($query->param) {
	#print STDERR "var=$var\n";
	if (my($stub,$rest) = $var =~ $self->_rx) {
	    my $field = $self->{'fields'}->{$stub};
	    #print STDERR "setnew stub=$stub rest=$rest val=",scalar($query->param($var)),"\n";
	    $field->setnew($rest, $query->param($var));
	}
    }
    my $method = $query->request_method;
    my %errors;
    while (my($fname,$field) = each %{$self->{'fields'}}) {
	# Tell field to look at setnew value(s)
	$field->endnew;
	if ($method eq 'POST' && $field->can('check')) {
	    my @err = $field->check;
	    $errors{$fname} = \@err if @err;
#	    print STDERR "$fname check failed\n" if @err;
	}
    }
    if ($method eq 'POST') {
	$self->{'errors'} = \%errors;
#	print STDERR "errors = ",%errors,"\n";
	return (%errors == 0);
    } else {
	return undef;
    }
}

#-----------------------------------------------------------------------
#	Get a regular expression matching field name and any extra.
#	E.g. if fields names are "a", "abc" and "def", the regexp
#	should match "a" "a/b", "abc/" returning repectively
#	("a",undef), ("a","b"), ("abc","") and not match "ab" etc.
#-----------------------------------------------------------------------
sub _rx {
    my $self = shift;
    unless ($self->{'rx'}) {
    # Create regexp matching all field names
	$self->{'rx'} = ("^(" .
			 join("|",
			      map {quotemeta($_)}
			      keys %{$self->{'fields'}}) .
			 ")(?:/(.*))?\$");
#	print STDERR "rx = $self->{'rx'}\n";
    }
    return $self->{'rx'};
}

#-----------------------------------------------------------------------
#	Return values
#-----------------------------------------------------------------------
sub values {
    my($self) = @_;
    my %values;
    while (my($fname,$field) = each %{$self->{'fields'}}) {
	$values{$fname} = $field->{'current'};
    }
    return \%values;
}

#-----------------------------------------------------------------------
#	Template-filling
#-----------------------------------------------------------------------
sub GET {
    my($self, $var) = @_;
    if (my($stub,$rest) = $var =~ $self->_rx) {
#	print STDERR "stub=$stub rest=$rest\n";
	if (defined(my $field = $self->{'fields'}->{$stub})) {
	    return $field->GET($rest);
	}
    }
    if ($var eq 'errors') {
	my @err = (map {{errortext => $_}}
		   map {@$_}
		   values %{$self->{'errors'}});
	return \@err;
    }
    return undef;
}

sub HTML {
    my($self, $var) = @_;
    if (my($stub,$rest) = $var =~ $self->_rx) {
	if (defined(my $field = $self->{'fields'}->{$stub})) {
	    if ($field->can('HTML')) {
		return $field->HTML($rest);
	    }
	}
    }
    return undef;
}

#=======================================================================
#	Text field
#	  initial => 'initial'
#	  validate => $re_or_code
#
#	{foo}
#=======================================================================
package PythonTech::Form::Text;
use strict;
use PythonTech::Conv qw(html_escape);

sub new {
    my $class = shift;
    my $self = {@_};
    bless $self, $class;
    $self->{'initial'} = '' unless defined $self->{'initial'};
    $self->{'current'} = $self->{'initial'};
    return $self;
}

sub setnew {
    my($self, $var,$value) = @_;
    if ($var eq '') {
	$self->{'new'} = $value;
    }
}

sub endnew {
    my $self = shift;
    if (defined $self->{'new'}) {
	$self->{'current'} = delete $self->{'new'};
    }
}

sub check {
    my $self = shift;
    my $validate = $self->{'validate'};
    if (! defined($validate)) {
	return;
    } elsif (ref($validate) eq 'CODE') {
	return $validate->($self->{'current'});
    } elsif (ref($validate) eq '') {
	# Assume regex
	if ($self->{'current'} !~ $validate) {
	    return ("Invalid format for $self->{'name'}");
	}
    }
    return;
}

sub GET {
    my($self, $var) = @_;
    if ($var eq '') {
	return $self->{'current'};
    }
}

sub HTML {
    my($self,$var) = @_;
    if ($var eq 'hidden') {
	my $hname = html_escape($self->{'name'});
	my $hvalue = html_escape($self->{'current'});
	return "<input type=\"hidden\" name=\"$hname\" value=\"$hvalue\" />";
    }
    return undef;
}

#=======================================================================
#	Date field
#	  initial => '2004-03-27'
#	  min => '1999-01-01'
#	  max => '2004-12-31'
#
#	{foo}				2005-04-13
#	{foo/dd} {foo/mm} {foo/yyyy}	13 04 2005
#	{foo/Mon} {foo/Month}		Apr April
#	{IF foo/dd/current/13}X{ENDIF}	X
#	{IF foo/mm/current/04}X{ENDIF}	X
#	{IF foo/yyyy/current/2005}X{ENDIF}	X
#	{foo/dd/selected/13}		 selected="selected"
#	{foo/mm/selected/12}
#	{foo/yyyy/selected/2005}	 selected="selected"
#	{foo/dd/checked/24}
#	{foo/mm/checked/04}		 checked="checked"
#	{foo/yyyy/checked/1999}
#	{foo/dd/option/13}		<option value="13" selected="selected">13</option>
#	{foo/mm/option/08}		<option value="08">08</option>
#	{foo/yyyy/option/2005}		<option value="2005" selected="selected">2005</option>
#	{foo/Mon/option/08}		<option value="08">Aug</option>
#	{foo/Month/option/1999}		<option value="1999">1999</option>
#	{foo/Month/menu}		<select name="foo/mm"><option...>
#=======================================================================
package PythonTech::Form::Date;
use strict;
use PythonTech::Conv qw(html_escape);

my @Month = qw(January February March April May June
	       July August September October November December);

sub new {
    my $class = shift;
    my $self = {@_};
    bless $self, $class;
    $self->{'initial'} = '' unless defined $self->{'initial'};
    $self->{'current'} = $self->{'initial'};
    return $self;
}

sub setnew {
    my($self, $var,$value) = @_;
    if (! defined $var) {
	$self->{'new'} = $value;
    } elsif ($var =~ /^(dd|mm|yyyy)$/) {
	$self->{"new/$var"} = $value;
    }
}

sub endnew {
    my $self = shift;
    my($yyyy,$mm,$dd,@errors);
    if (defined $self->{'new'}) {
	unless (($yyyy,$mm,$dd) = 
		$self->{'new'} =~ /^(\d\d\d\d)-(\d\d)-(\d\d)$/) {
	    push @errors, "Invalid date format";
	}
    } else {
	($dd) = $self->{'new/dd'} =~ /^(\d\d)$/ or
	    push @errors, "Invalid day '$self->{'new/dd'}'";
	($mm) = $self->{'new/mm'} =~ /^(\d\d)$/ or
	    push @errors, "Invalid month '$self->{'new/mm'}'";
	($yyyy) = $self->{'new/yyyy'} =~ /^(\d\d\d\d)$/ or
	    push @errors, "Invalid year '$self->{'new/yyyy'}'";
	unless (@errors) {
	    $self->{'new'} = "$yyyy-$mm-$dd";
	}
    }
    unless (@errors) {
	push @errors, "Day out of range"
	    if $dd < 1 || $dd > 31;
	push @errors, "Month out of range"
	    if $mm < 1 | $mm > 12;
    }
    unless (@errors) {
	push @errors, "Date before earliest allowed"
	    if defined $self->{'min'} && $self->{'new'} lt $self->{'min'};
	push @errors, "Date after latest allowed"
	    if defined $self->{'min'} && $self->{'new'} gt $self->{'max'};
    }
    if ($self->{'new'}) {
	$self->{'current'} = delete $self->{'new'};
    }
    foreach my $item (grep {m!^new/.*!} keys %$self) {
	delete $self->{$item};
    }
    $self->{'errors'} = \@errors;
}

sub check {
    my $self = shift;
    return @{$self->{'errors'}};
}

sub GET {
    my($self, $var) = @_;
    my($yyyy,$mm,$dd) = $self->{'current'} =~ /^(\d\d\d\d)-(\d\d)-(\d\d)$/;
    if (! defined $var) {
	return $self->{'current'};
    } elsif ($var eq 'dd') {
	return $dd;
    } elsif ($var eq 'mm') {
	return $dd;
    } elsif ($var eq 'yyyy') {
	return $yyyy;
    } elsif ($var eq 'Mon') {
	return substr($Month[$mm-1],0,3) || 'Xxx';
    } elsif ($var eq 'Month') {
	return $Month[$mm-1] || 'Unknown';
    } elsif (my($qdd) = $var =~ m!^dd/current/(\d\d)$!) {
	return $qdd eq $dd ? 1 : 0;
    } elsif (my($qmm) = $var =~ m!^mm/current/(\d\d)$!) {
	return $qmm eq $mm ? 1 : 0;
    } elsif (my($qyyyy) = $var =~ m!^yyyy/current/(\d\d\d\d)$!) {
	return $qyyyy eq $yyyy ? 1 : 0;
    }
    return undef;
}

sub HTML {
    my($self, $var) = @_;
    my($yyyy,$mm,$dd) = $self->{'current'} =~ /^(\d\d\d\d)-(\d\d)-(\d\d)$/;
    my $hname = html_escape($self->{'name'});
    if (my($qdd) = $var =~ m!^dd/selected/(\d\d)$!) {
	return $qdd eq $dd ? ' selected="selected"' : '';
    } elsif (my($qmm) = $var =~ m!^mm/selected/(\d\d)$!) {
	return $qmm eq $mm ? ' selected="selected"' : '';
    } elsif (my($qyyyy) = $var =~ m!^yyyy/selected/(\d\d\d\d)$!) {
	return $qyyyy eq $yyyy ? ' selected="selected"' : '';
    } elsif (my($qdd) = $var =~ m!^dd/checked/(\d\d)$!) {
	return $qdd eq $dd ? ' checked="checked"' : '';
    } elsif (my($qmm) = $var =~ m!^mm/checked/(\d\d)$!) {
	return $qmm eq $mm ? ' checked="checked"' : '';
    } elsif (my($qyyyy) = $var =~ m!^yyyy/checked/(\d\d\d\d)$!) {
	return $qyyyy eq $yyyy ? ' checked="checked"' : '';
    } elsif (my($qdd) = $var =~ m!^dd/option/(\d\d)$!) {
	my $sel = $qdd eq $dd ? ' selected="selected"' : '';
	return "<option value=\"$qdd\"$sel>$qdd</option>";
    } elsif (my($qmm) = $var =~ m!^mm/option/(\d\d)$!) {
	my $sel = $qmm eq $mm ? ' selected="selected"' : '';
	return "<option value=\"$qmm\"$sel>$qmm</option>";
    } elsif (my($qyyyy) = $var =~ m!^yyyy/option/(\d\d\d\d)$!) {
	my $sel = $qyyyy eq $yyyy ? ' selected="selected"' : '';
	return "<option value=\"$qyyyy\"$sel>$qyyyy</option>";
    } elsif (my($qmm) = $var =~ m!^Mon/option/(\d\d)$!) {
	my $sel = $qmm eq $mm ? ' selected="selected"' : '';
	return "<option value=\"$qmm\"$sel>" . 
	    substr($Month[$qmm-1],0,3) .
		"</option>";
    } elsif (my($qmm) = $var =~ m!^Month/option/(\d\d)$!) {
	my $sel = $qmm eq $mm ? ' selected="selected"' : '';
	return "<option value=\"$qmm\"$sel>" . 
	    $Month[$qmm-1] .
		"</option>";
    } elsif (my($style) = $var =~ m!^(dd)/menu$!) {
	my $hname = html_escape($self->{'name'});
	my $html = "<select name=\"$hname/dd\">";
	foreach my $d (1..31) {
	    my $qdd = sprintf("%02d",$d);
	    $html .= $self->HTML("$style/option/$qdd");
	}
	$html .= "</select>";
	return $html;
    } elsif (my($style) = $var =~ m!^(mm|Mon|Month)/menu$!) {
	my $hname = html_escape($self->{'name'});
	my $html = "<select name=\"$hname/mm\">";
	foreach my $m (1..12) {
	    my $qmm = sprintf("%02d",$m);
	    $html .= $self->HTML("$style/option/$qmm");
	}
	$html .= "</select>";
	return $html;
    } elsif ($var eq 'hidden') {
	my $hname = html_escape($self->{'name'});
	my $hvalue = html_escape($self->{'current'});
	return "<input type=\"hidden\" name=\"$hname\" value=\"$hvalue\" />";
    }
    return undef;
}

#=======================================================================
#	Select object from list
#	  list => [@objist],
#	  show => $attr_or_code_or_hash,
#	  initial => $obj,
#
#	{LOOP foo}
#	 {IF _/current}X{ENDIF}		X
#	 {_/selected}			 selected="selected"
#	{ENDLOOP}
#	{foo/menu}			<select name="foo"><option ...>
#	{foo/options}
#=======================================================================
package PythonTech::Form::SelectObj;
use strict;
use PythonTech::Conv qw(html_escape);

sub new {
    my $class = shift;
    my $self = {@_};
    bless $self, $class;
    my @list;
    foreach my $obj (@{$self->{'list'}}) {
	push @list, PythonTech::Form::SelectObj::Option->new(parent => $self,
							     obj => $obj);
    }
    $self->{'options'} = \@list;
    $self->{'current'} = $self->{'initial'};
    return $self;
}

sub setnew {
    my($self, $var,$value) = @_;
    if (! defined $var) {
	# Search for object with given uri
	my @match = grep {$_->uri eq $value} @{$self->{'list'}};
	$self->{'new'} = $match[0]; # undef if not in list
    }
}

sub endnew {
    my $self = shift;
    if (exists $self->{'new'}) {
	$self->{'current'} = delete $self->{'new'};
    }
}

sub check {
    my $self = shift;
    my @errors;
    unless ($self->{'current'}) {
	push @errors, "Unknown id";
    }
    return @errors;
}

sub GET {
    my($self, $var) = @_;
    if ($var eq '') {
	# List of available options
	return $self->{'options'};
    }
    return undef;
}

sub HTML {
    my($self, $var) = @_;
    if ($var eq 'menu') {
	my $hname = html_escape($self->{'name'});
	my $html = "<select name=\"$hname\">";
	$html .= $self->HTML('options');
	$html .= "</select>";
	return $html;
    } elsif ($var eq 'options') {
	my $html = "";
	foreach my $opt (@{$self->{'options'}}) {
	    $html .= $opt->HTML("_/option");
	}
	return $html;
    } elsif ($var eq 'hidden') {
	my $hname = html_escape($self->{'name'});
	my $cur = $self->{'current'};
	my $hvalue = $cur ? html_escape($cur->uri) : '';
	return "<input type=\"hidden\" name=\"$hname\" value=\"$hvalue\" />";
    }
    return undef;
}

#=======================================================================
#	Option from object list
#	  parent => SelectObj
#	  obj => $obj
#
#	{/id}				baseuri/myid
#	{/show}				My Name
#	{IF /current}X{ENDIF}		X
#	{/selected}			 selected="selected"
#	{/checked}			 checked="checked"
#	{/option/baseuri/myid}		<option value="baseuri/myid" selected="selected">My Name</option>
#=======================================================================
package PythonTech::Form::SelectObj::Option;
use strict;
use PythonTech::Conv qw(html_escape);

sub new {
    my $class = shift;
    my $self = {@_};
    bless $self, $class;
    return $self;
}

sub show {
    my $self = shift;
    my $show = $self->{'parent'}->{'show'};
    my $obj = $self->{'obj'};
    if (! defined $show) {
	# Fall through to default
    } elsif (ref($show) eq '') {
	# Scalar - assume attribute name
	my $attrval = $obj->attr($show);
	return $attrval;
    } elsif (ref($show) eq 'CODE') {
	# Subroutine to deliver display name
	my $text = $show->($obj);
	return $text;
    } elsif (ref($show) eq 'HASH') {
	my $value = $show->{$obj->uri};
	return $value if defined $value;
    }
    # Default: return trailing word-characters from URI
    (my $tail = $obj->uri) =~ s!.*\W!!;
    return $tail;
}

sub GET {
    my($self, $var) = @_;
    my $obj = $self->{'obj'};
    if ($var eq '_/id') {
	return $obj->uri;
    } elsif ($var eq '_/show') {
	return $self->show;
    } elsif ($var eq '_/current') {
	return ($self->{'parent'}->{'current'} eq $obj) ?
	    1 : 0;
    }
    return undef;
}

sub HTML {
    my($self, $var) = @_;
    my $obj = $self->{'obj'};
    if ($var eq '_/selected') {
	#print STDERR "current=$self->{'parent'}->{'current'} obj=$obj\n";
	return ($self->{'parent'}->{'current'} eq $obj) ?
	    ' selected="selected"' : '';
    } elsif ($var eq '_/checked') {
	return ($self->{'parent'}->{'current'} eq $obj) ?
	    ' checked="checked"' : '';
    } elsif ($var eq '_/option') {
	my $hid = html_escape($obj->uri);
	my $sel = ($self->{'parent'}->{'current'} eq $obj) ?
	    ' selected="selected"' : '';
	my $hshow = html_escape($self->show);
	return "<option value=\"$hid\"$sel>$hshow</option>";
    } elsif ($var eq '_/radio') {
	my $hname = html_escape($self->{'parent'}->{'name'});
	my $hid = html_escape($obj->uri);
	my $chk = ($self->{'parent'}->{'current'} eq $obj) ?
	    ' checked="checked"' : '';
	return "<input type=\"radio\" name=\"$hname\" value=\"$hid\"$chk />";
    }
    return undef;
}

1;
